#pragma once
#include <string>
#include <memory>

#include "headers.h"
#include "esphome/core/optional.h"

#ifdef USE_SOCKET_IMPL_LWIP_TCP
#if LWIP_IPV6
#define USE_SOCKET_IPV6
#endif
#endif

namespace esphome {
namespace socket {

class Socket {
 public:
  Socket() = default;
  virtual ~Socket() = default;
  Socket(const Socket &) = delete;
  Socket &operator=(const Socket &) = delete;

  virtual std::unique_ptr<Socket> accept(struct sockaddr *addr, socklen_t *addrlen) = 0;
  virtual int bind(const struct sockaddr *addr, socklen_t addrlen) = 0;
  virtual int connect(const struct sockaddr *addr, socklen_t addrlen) = 0;
  /**
   * @brief Helper to check if a socket connect() that was EINPROGRESS is now finished.
   *
   * If the connect finnished successfully, returns 0.
   * If it's still in progress, returns -1 and sets errno to EINPROGRESS.
   * Other errors result in return code -1 and errno like in blocking connect().
   */
  virtual int connect_finished() = 0;

  virtual int close() = 0;
  virtual int shutdown(int how) = 0;

  virtual int getpeername(struct sockaddr *addr, socklen_t *addrlen) = 0;
  virtual std::string getpeername() = 0;
  virtual int getsockname(struct sockaddr *addr, socklen_t *addrlen) = 0;
  virtual std::string getsockname() = 0;
  virtual int getsockopt(int level, int optname, void *optval, socklen_t *optlen) = 0;
  virtual int setsockopt(int level, int optname, const void *optval, socklen_t optlen) = 0;
  virtual int listen(int backlog) = 0;
  virtual ssize_t read(void *buf, size_t len) = 0;
  virtual ssize_t readv(const struct iovec *iov, int iovcnt) = 0;
  virtual ssize_t write(const void *buf, size_t len) = 0;
  virtual ssize_t writev(const struct iovec *iov, int iovcnt) = 0;
  virtual int setblocking(bool blocking) = 0;
  virtual int loop() { return 0; };
};

/// Create a socket of the given domain, type and protocol.
std::unique_ptr<Socket> socket(int domain, int type, int protocol);

/// Create a socket in the newest available IP domain (IPv6 or IPv4) of the given type and protocol.
std::unique_ptr<Socket> socket_ip(int type, int protocol);

/// Set a sockaddr to the any address for the IP version used by socket_ip().
socklen_t set_sockaddr_any(struct sockaddr *addr, socklen_t addrlen, uint16_t port);

}  // namespace socket
}  // namespace esphome
